from abc import ABC, abstractmethod
from typing import Callable
from logging import getLogger
from collections import deque

logger = getLogger(__name__)

class Event(ABC):
    """
    Event is an abstract base class for domain events.
    """

class Command(ABC):
    """
    Command is a class representing a request to perform an action.
    """
    ...
    @abstractmethod
    def execute(self):
        """
        Executes the command.
        """
        ...

class Messagebus:
    """
    Messagebus is a class that routes domain events and commands to their respective handlers.

    Attributes:
        handlers: A dictionary mapping command types to their corresponding handlers.
        consumers: A dictionary mapping event types to a list of their consumers.
    """

    def __init__(self):
        self.handlers = dict[type[Command], Callable[[Command], None]]()
        self.consumers = dict[type[Event], list[Callable[[Event], None]]]()
        self.queue = deque[Command | Event]()

    def register(self, command_type: type[Command], handler: Callable[[Command], None]):
        """
        Sets a handler for a given command type. A command type can only have one handler.
        Parameters:
            command_type: The type of the command.
            handler: The handler to be registered.
        """
        self.handlers[command_type] = handler

    def subscribe(self, event_type: type[Event], consumer: Callable[[Event], None]):
        """
        Adds a consumer for a given event type. An event type can have multiple consumers.
        Parameters:
            event_type: The type of the event.
            consumer: The consumer to be added.
        """
        self.consumers.setdefault(event_type, []).append(consumer)

    def handle(self, command: Command):
        """
        Handles a given command by invoking its corresponding handler 
        or executing it by default.

        Parameters:
            command: The command to be handled.
        """
        handler = self.handlers.get(type(command), None)
        command.execute() if not handler else handler(command)

    def consume(self, event: Event):
        """
        Consumes a given event by invoking its registered consumers.

        Parameters:
            event: The event to be consumed.
        """
        for consumer in self.consumers.get(type(event), []):
            try:
                consumer(event)
            except Exception as exception:
                logger.error(f"Error while consuming event {event}")
                logger.debug(exception, exc_info=True)


from pybondi.messagebus import Event
from pybondi.messagebus import Command
from pybondi.messagebus import Messagebus
from pybondi.aggregate import Aggregate
from pybondi.aggregate import Root
from pybondi.aggregate import Factory
from pybondi.session import Session
from pybondi.publisher import Publisher
from pybondi.repository import Repository
from pybondi.callbacks import Callbacks


from abc import ABC, abstractmethod
from typing import Any
from collections import deque
from pybondi.messagebus import Event

class Root:
    '''
    Root is a class that represents the root of an aggregate. It is responsible for
    mantaning the consistency of the aggregate by storing the events that have occurred

    Attributes:
        id: The unique identifier of the aggregate.
        events: A deque storing the events that have occurred on the aggregate.
    '''

    def __init__(self, id: Any):
        self.id = id
        self.events = deque[Event]()

    def publish(self, event: Event):
        assert isinstance(event, Event), 'The event must be an instance of Event'
        '''
        Publishes a domain event to the aggregate.
        Parameters:
            event: The event to be published.
        '''
        self.events.append(event)


class Aggregate(ABC):
    '''
    Aggregate is an abstract class that represents a collection of domain objects that are
    treated as a single unit. It is responsible for maintaining the consistency of the domain
    objects by enforcing invariants and ensuring that all changes are made through 
    well-defined operations.

    Attributes:
        root: The root of the aggregate.
    '''        
    def __init__(self, root: Root):
        self.root = root


class Factory(ABC):
    '''
    Factory is an abstract class that defines the interface for creating aggregates.
    Aggregates should be created through a factory to ensure that they are properly
    initialized.    
    '''

    @abstractmethod
    def __call__(self, *args, **kwargs) -> Aggregate:
        '''
        Creates a new aggregate.
        '''
        ...


from collections import deque
from typing import Callable

from pybondi.aggregate import Aggregate
from pybondi.messagebus import Messagebus, Command, Event
from pybondi.repository import Repository
from pybondi.publisher import Publisher

class Session:
    """
    A session manages a unit of work, cordinates the repository, the message bus and
    the publisher, mantaing the transactional boundaries.
    """
    
    event_handlers = dict[type[Event], list[Callable[[Event], None]]]()
    command_handlers = dict[type[Command], Callable[[Command], None]]()

    def __init__(self, repository: Repository = None, publisher: Publisher = None, messagebus: Messagebus = None):
        self.publisher = publisher or Publisher()
        self.repository = repository or Repository()
        self.queue = deque[Command | Event]()

        if messagebus:
            self.messagebus = messagebus
        else:
            self.messagebus = Messagebus()
            for event_type, handlers in self.event_handlers.items():
                [self.messagebus.subscribe(event_type, handler) for handler in handlers]
            for command_type, handler in self.command_handlers.items():
                self.messagebus.register(command_type, handler)

    def enqueue(self, message: Command | Event):
        """
        Enqueues a message in the session queue.
        """
        self.queue.append(message)

    def dequeue(self) -> Command | Event:
        """
        Dequeues a message from the session queue.
        """
        return self.queue.popleft()

    def add(self, aggregate: Aggregate):
        """
        Adds an aggregate to the repository.
        """
        self.repository.add(aggregate)

    def dispatch(self, message: Command | Event):
        """
        Dispatches a message to the message bus.
        """
        if isinstance(message, Command):
            self.messagebus.handle(message)
        elif isinstance(message, Event):
            self.messagebus.consume(message)

    def run(self):
        """
        Processes all messages in the queue.
        """
        while self.queue:
            message = self.dequeue()
            self.dispatch(message)
            
            for aggregate in self.repository.aggregates.values():
                while aggregate.root.events:
                    event = aggregate.root.events.popleft()
                    self.enqueue(event)


    def execute(self, command: Command):
        """
        Executes a command by enqueuing it in the message bus and processing all messages in the queue.
        """
        self.enqueue(command)
        self.run()

    def begin(self):
        """
        Begins a new transaction.
        """
        self.publisher.begin()

    def commit(self):
        """
        Commits changes from the transaction.
        """
        self.repository.commit(), self.publisher.commit()

    def rollback(self):
        """
        Rolls back changes of the transaction.
        """
        self.repository.rollback(), self.publisher.rollback()

    def close(self):
        """
        Closes the session. If the session queue is not empty, raises an exception.
        """
        self.repository.close(), self.publisher.close()
        if self.queue:
            raise Exception("Session queue is not empty")

    def __enter__(self):
        self.publisher.begin()
        return self
    
    def __exit__(self, exc_type, exc_value, traceback):
        self.run()
        if exc_type:
            self.rollback()
        else:
            self.commit()
        self.close()



from typing import Any
from typing import Callable
from collections import deque
from abc import ABC, abstractmethod

class Base(ABC):
    """
    An abstract base class for subscribers to a publisher.

    Subscribers are responsible for receiving messages, managing transactions,
    and closing their connections.
    """

    @abstractmethod
    def handle(self,topic: str, message: Any) -> None:
        """
        Receives a message from the publisher.

        Args:
            message: The message to be received.
        """

    @abstractmethod
    def begin(self) -> None:
        """
        Starts a new transaction.
        """

    @abstractmethod
    def commit(self) -> None:
        """
        Commits the current transaction.
        """

    @abstractmethod
    def rollback(self) -> None:
        """
        Rolls back the current transaction.
        """

    @abstractmethod
    def close(self) -> None:
        """
        Closes the subscriber and its connection.
        """

class Subscriber(Base):
    def __init__(self):
        self.handlers = dict[str, list[Callable]]()
        self.queue = deque[tuple[str, Any]]()

    def subscribe(self, topic: str, handler: Callable):
        self.handlers.setdefault(topic, []).append(handler)

    def handle(self, topic: str, message: Any):
        self.queue.append((topic, message))

    def commit(self):
        while self.queue:
            topic, message = self.queue.popleft()
            [handler(message) for handler in self.handlers.get(topic, [])]

    def rollback(self):
        self.queue.clear()

    def begin(self):
        [handler('begin', None) for handler in self.handlers.get('begin', [])]

    def close(self):
        [handler('begin', None) for handler in self.handlers.get('close', [])]


class Publisher:
    """
    A publisher that manages a set of subscribers and publishes messages to them.

    The publisher is responsible for subscribing and unsubscribing subscribers,
    publishing messages to subscribed topics, and coordinating transactions
    across all subscribers.

    The difference between a publisher and a message bus is that a publisher
    is responsible for publish data from the inside of the bounded context
    to outside systems. A message bus is responsible for routing events and
    commands within the bounded context.
    """

    def __init__(self) -> None:
        """
        Initializes the publisher with an empty subscriber dictionary.
        """
        self.subscribers = list[Subscriber]()

    def subscribe(self, subscriber: Subscriber) -> None:
        """
        Subscribes a new subscriber to the publisher.

        Parameters:
            subscriber: The subscriber to subscribe.
        """
        assert isinstance(subscriber, Base), 'Subscriber should inherit from Base'
        self.subscribers.append(subscriber)

    def publish(self, topic: str, message: Any) -> None:
        """
        Publishes a message to a specific topic.

        The message is sent to all subscribers of the given topic.

        Parameters:
            topic: The topic to publish to.
            message: The message to be published.
        """
        [subscriber.handle(topic, message) for subscriber in self.subscribers]

    def commit(self) -> None:
        """
        Commits the current transaction for all subscribers.
        """
        [subscriber.commit() for subscriber in self.subscribers]
            

    def rollback(self) -> None:
        """
        Rolls back the current transaction for all subscribers.
        """
        [subscriber.rollback() for subscriber in self.subscribers]


    def begin(self) -> None:
        """
        Starts a new transaction for all subscribers.
        """
        [subscriber.begin() for subscriber in self.subscribers]

    def close(self) -> None:
        """
        Closes all subscribers and their connections.
        """
        [subscriber.close() for subscriber in self.subscribers]